from functools import partial
import pandas as pd
import numpy as np
from datacode.summarize import format_numbers_to_decimal_places
from datacode.formatters.stars import convert_to_stars
from datacode.psm.names import matched_var, t_stat_var, diff_var
from datacode.psm.typing import StrList



def matching_summary_stats(df: pd.DataFrame, matched_df: pd.DataFrame,
                           treated_var: str, describe_vars: StrList,
                           entity_var: str,
                           control_name: str = 'Control', treated_name: str = 'Treated') -> pd.DataFrame:
    common_args = (
        treated_var,
        describe_vars,
        entity_var
    )

    common_kwargs = dict(
        treated_name=treated_name
    )

    summ = mean_and_diff_df_by_treatment(
        matched_df,
        *common_args,
        control_name=matched_var,
        **common_kwargs
    )

    control_summ = mean_and_diff_df_by_treatment(
        df,
        *common_args,
        control_name=control_name,
        **common_kwargs
    )

    summ[control_name] = control_summ[control_name]
    summ = summ[[control_name, treated_name, matched_var, diff_var, t_stat_var]]

    return summ


def mean_and_diff_df_by_treatment(df: pd.DataFrame, treated_var: str, describe_vars: StrList,
                                  entity_var: str, num_decimals: int = 2, coerce_ints: bool = True,
                                  control_name: str = 'Control', treated_name: str = 'Treated') -> pd.DataFrame:
    common_args = (
        treated_var,
        describe_vars,
        entity_var
    )

    common_kwargs = dict(
        treated_name=treated_name,
        control_name=control_name
    )

    mean_df = _stat_df_by_treatment(
        df,
        *common_args,
        stat='mean',
        **common_kwargs
    )

    std_df = _stat_df_by_treatment(
        df,
        *common_args,
        stat='std',
        **common_kwargs
    )

    count_df = _stat_df_by_treatment(
        df,
        *common_args,
        stat='count',
        **common_kwargs
    )

    diff_t = _diff_and_t_df_from_mean_and_std_df(
        mean_df,
        std_df,
        count_df,
        control_name=control_name,
        treated_name=treated_name
    )

    summ = pd.concat([mean_df, diff_t], axis=1)
    summ = _format_summary_df(summ, num_decimals=num_decimals, coerce_ints=coerce_ints)

    return summ


def _format_summary_df(df: pd.DataFrame, float_format: str = '.2f', num_decimals: int = 2,
                       coerce_ints: bool = True,) -> pd.DataFrame:
    #     num_formatter = lambda x: f'{x:{float_format}}' if isinstance(x, float) else x
    num_formatter = partial(format_numbers_to_decimal_places, decimals=num_decimals, coerce_ints=coerce_ints)
    df[diff_var] = df[diff_var].apply(num_formatter)
    df[diff_var] = df[diff_var] + df[t_stat_var].apply(
        lambda x: convert_to_stars(x) if isinstance(x, float) else x
    )
    df = df.applymap(num_formatter)

    return df


def _stat_df_by_treatment(df: pd.DataFrame, treated_var: str, describe_vars: StrList, entity_var: str ,
                          stat: str = 'mean',
                          control_name: str = 'Control', treated_name: str = 'Treated') -> pd.DataFrame:
    grouped = df.groupby(treated_var)[describe_vars]
    agg_func = getattr(grouped, stat)
    stat_df = agg_func().T
    stat_df.columns = [control_name, treated_name]
    count_series = _count_series(
        df,
        treated_var,
        entity_var,
        control_name=control_name,
        treated_name=treated_name
    )
    stat_df = stat_df.append(count_series)

    return stat_df

def _diff_and_t_df_from_mean_and_std_df(mean_df: pd.DataFrame, std_df: pd.DataFrame, count_df: pd.DataFrame,
                                        control_name: str = 'Control',
                                        treated_name: str = 'Treated') -> pd.DataFrame:
    diff = mean_df[treated_name] - mean_df[control_name]
    standard_errors = (
                              (std_df[control_name] ** 2 ) /count_df[control_name] +
                              (std_df[treated_name] ** 2 ) /count_df[treated_name]
                      ) ** 0.5
    t_stats = diff /standard_errors
    t_stats = t_stats.replace(np.inf, '')

    df = pd.DataFrame([diff, t_stats]).T
    colnames = [diff_var, t_stat_var]
    df.columns = colnames
    df.loc['N', colnames] = ''

    return df

def _count_series(df: pd.DataFrame, treated_var: str, entity_var: str,
                  control_name: str = 'Control', treated_name: str = 'Treated') -> pd.Series:
    count_s = df.groupby(treated_var)[entity_var].count()
    count_s.index = [control_name, treated_name]
    count_s.name = 'N'

    return count_s


import pandas as pd

from datacode.ff.create.model import parse_model
from datacode.ff.fftypes import StrOrInt, TwoStrTupleList, TwoStrTuple

def create_dual_sort_variables_get_pairings(df: pd.DataFrame, factor_model: StrOrInt, size_var: str=None, value_var: str=None,
                               profitability_var: str=None, investment_var: str=None,
                               custom_pairings: TwoStrTupleList=None) -> TwoStrTupleList:

    pairings = _select_pairings(
        factor_model=factor_model,
        size_var=size_var,
        value_var=value_var,
        profitability_var=profitability_var,
        investment_var=investment_var,
        custom_pairings=custom_pairings
    )

    _create_dual_sort_variables(df, pairings)

    return pairings

def _create_dual_sort_variables(df: pd.DataFrame, parings: TwoStrTupleList) -> None:
    [_create_dual_sort_variable(df, pairing) for pairing in parings]

def _create_dual_sort_variable(df: pd.DataFrame, pairing: TwoStrTuple) -> None:
    combined_name = _dual_sort_varname(pairing[0], pairing[1])
    df[combined_name] = df[pairing[0]].astype(str) + df[pairing[1]].astype(str)


def _select_pairings(factor_model: StrOrInt, size_var: str=None, value_var: str=None,
                     profitability_var: str=None, investment_var: str=None,
                     custom_pairings: TwoStrTupleList=None) -> TwoStrTupleList:

    factor_model = parse_model(factor_model)

    if factor_model == 'custom':
        return custom_pairings

    if factor_model == 3:
        return [
            (size_var, value_var)
        ]

    if factor_model == 5:
        return [
            (size_var, value_var),
            (size_var, profitability_var),
            (size_var, investment_var)
        ]


def _dual_sort_varname(portvar1: str, portvar2: str) -> str:
    return f'{portvar1}/{portvar2}'


import datetime
from copy import deepcopy
from typing import Union, Optional, Type

import pandas as pd

from datacode.models.dtypes.base import DataType


class PeriodType(DataType):
    name_roots = ('period',)

    def __init__(self, freq: str, categorical: bool = False, ordered: bool = False):
        super().__init__(
            datetime.datetime,
            pd_class=pd.PeriodDtype,
            categorical=categorical,
            ordered=ordered,
        )
        self.freq = freq
        self.equal_attrs = deepcopy(self.equal_attrs)
        self.equal_attrs.append('freq')
        self.repr_cols = deepcopy(self.repr_cols)
        self.repr_cols.append('freq')

    @classmethod
    def from_str(cls, dtype: str, categorical: bool = False, ordered: bool = False):
        dtype = dtype.lower()
        freq: Optional[str] = None
        for name in cls.name_roots:
            if dtype.startswith(name):
                _, freq_extra = dtype.split('[')
                freq = freq_extra.strip(']')
        if freq is None:
            raise ValueError(f'Dtype {dtype} does not match valid names for {cls.__name__}: {cls.names}')
        return cls(
            freq,
            categorical=categorical,
            ordered=ordered
        )

    @property
    def read_file_arg(self) -> Union[Type, str]:
        return f'period[{self.freq}]'



from typing import List, Optional
import pandas as pd
from datacode.panel.did.dataprep import diff_df_from_panel_df
from datacode.panel.did.reg import diff_reg_for_each_yvar_and_produce_summary
from datacode.panel.did.totex import diff_reg_summary_to_latex_table_and_output
import pyexlatex.table as lt

def diff_reg_for_each_yvar_from_panel_df_latex_table(df: pd.DataFrame, entity_var: str, time_var: str, treated_var: str,
                                                     yvars: List[str], treated_time_var: str = 'After',
                                                     xvars: Optional[List[str]] = None,
                                                     caption: str = 'Difference-in-Difference Regressions',
                                                     extraname: Optional[str] = None,
                                                     below_text: Optional[str] = None,
                                                     outfolder: Optional[str] = None) -> lt.Table:
    reg_list, summ = diff_reg_for_each_yvar_from_panel_df(
        df,
        entity_var,
        time_var,
        treated_var,
        yvars,
        treated_time_var=treated_time_var,
        xvars=xvars
    )

    table = diff_reg_summary_to_latex_table_and_output(
        summ,
        entity_var,
        caption=caption,
        extraname=extraname,
        below_text=below_text,
        outfolder=outfolder,
        xvars=xvars
    )

    return table

def diff_reg_for_each_yvar_from_panel_df(df: pd.DataFrame, entity_var: str, time_var: str, treated_var: str,
                                         yvars: List[str], treated_time_var: str = 'After',
                                         xvars: Optional[List[str]] = None):
    diff_df = diff_df_from_panel_df(
        df,
        entity_var,
        time_var,
        treated_var,
        treated_time_var=treated_time_var
    )

    reg_list, summ = diff_reg_for_each_yvar_and_produce_summary(
        diff_df,
        yvars,
        treated_var,
        treated_time_var=treated_time_var,
        xvars=xvars
    )

    return reg_list, summ


from copy import deepcopy
from typing import Type, Union
import re

import numpy as np
import pandas as pd

from datacode.models.dtypes.base import DataType
from datacode.models.dtypes.bit_size import get_bit_from_dtype


class IntType(DataType):
    names = ('int', 'integer', 'int_', 'int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64')

    def __init__(self, categorical: bool = False, ordered: bool = False, bit_size: int = 64):
        self.bit_size = bit_size
        super().__init__(
            int,
            pd_class=self._get_pd_class(),
            categorical=categorical,
            ordered=ordered,
            is_numeric=True
        )
        self.equal_attrs = deepcopy(self.equal_attrs)
        self.equal_attrs.append('bit_size')
        self.repr_cols = deepcopy(self.repr_cols)
        self.repr_cols.append('bit_size')

    def _get_pd_class(self) -> Type:
        if self.bit_size == 8:
            return pd.Int8Dtype
        elif self.bit_size == 16:
            return pd.Int16Dtype
        elif self.bit_size == 32:
            return pd.Int32Dtype
        elif self.bit_size == 64:
            return pd.Int64Dtype
        else:
            raise ValueError(f'must pass bit_size of 8, 16, 32, or 64. Got {self.bit_size}')

    @property
    def read_file_arg(self) -> Union[Type, str]:
        """
        Must use string for nullable int type loading
        :return:
        """
        if self.bit_size == 8:
            return 'Int8'
        elif self.bit_size == 16:
            return 'Int16'
        elif self.bit_size == 32:
            return 'Int32'
        elif self.bit_size == 64:
            return 'Int64'
        else:
            raise ValueError(f'must pass bit_size of 8, 16, 32, or 64. Got {self.bit_size}')

    @classmethod
    def from_str(cls, dtype: str, categorical: bool = False, ordered: bool = False):
        dtype = dtype.lower()
        if dtype not in cls.names:
            raise ValueError(f'Dtype {dtype} does not match valid names for {cls.__name__}: {cls.names}')
        bit_or_none = get_bit_from_dtype(dtype)
        if bit_or_none is None:
            bit_or_none = 64
        return cls(
            categorical=categorical,
            ordered=ordered,
            bit_size=bit_or_none
        )

    @property
    def index_arg(self) -> Union[Type, str]:
        return pd.Int64Index



from .base import NamedList, NamedSet


class PipelineOptions:

    def __init__(self, data_sources=False, digraph=True, node_kwargs={}, graph_kwargs={}):
        self.data_sources = data_sources
        self.digraph = digraph
        self.node_attr = node_kwargs
        self.graph_attr = graph_kwargs

    def keys(self):
        return ['data_sources', 'digraph', 'node_attr', 'graph_attr']

    def __getitem__(self, item):
        return getattr(self, item)

    def __repr__(self):
        return f'<PipelineOptions(data_sources={self.data_sources}, digraph={self.digraph})>'


class IDSet(NamedSet):
    pass

class IDCollection(NamedList):
    pass

    def to_id_list(self):
        out_set = set()
        for set_ in self:
            out_set.update(set_)
        return list(out_set)


class MatchData:

    def __iter__(self):
        return (i for i in self.tuple)

    def __len__(self):
        return len(self.tuple)

    def __getitem__(self, item):
        return self.tuple[item]


class IDComparison(MatchData):

    def __init__(self, left_unmatched, matched, right_unmatched, name=None):
        self.left_unmatched = left_unmatched
        self.matched = matched
        self.right_unmatched = right_unmatched
        self.name = name
        self.tuple = (left_unmatched, matched, right_unmatched)


class MatchComparisonBarData(IDComparison):

    def __init__(self, left_unmatched, matched, right_unmatched, name=None):
        super().__init__(left_unmatched, matched, right_unmatched, name=name)
        self._set_midpoints()

    def _set_midpoints(self):
        left_midpoint = self.left_unmatched/2
        matched_midpoint = self.left_unmatched + self.matched/2
        right_midpoint = self.left_unmatched + self.matched + self.right_unmatched/2
        self.midpoints = (left_midpoint, matched_midpoint, right_midpoint)

    @classmethod
    def from_id_comparison(cls, id_comp: IDComparison):
        return cls(*id_comp, name=id_comp.name)


class IDComparisonCollection(NamedList):
    pass


class MatchComparisonBarGraphData(MatchData):

    def __init__(self, left_unmatched_tup, matched_tup, right_unmatched_tup, name=None):
        self.left_unmatched_tup = left_unmatched_tup
        self.matched_tup = matched_tup
        self.right_unmatched_tup = right_unmatched_tup
        self.name = name
        self.tuple = (left_unmatched_tup, matched_tup, right_unmatched_tup)


import pandas as pd

from pd_utils.pdutils import _to_list_if_str
from datacode.ff.fftypes import (
    ListOrStr,
    StrOrInt,
    DictofStrsandStrLists,
    GroupvarNgroupsDict,
    TwoStrTupleList,
    TwoStrTuple,
    StrBoolDict
)
from datacode.ff.create.model import _validate_model
from datacode.ff.create.sort import create_ff_portfolios, _other_groupvar_portname
from datacode.ff.create.label import get_and_set_labels
from datacode.ff.create.dualsort import create_dual_sort_variables_get_pairings, _dual_sort_varname
from datacode.ff.create.average import portfolio_returns, market_returns
from datacode.ff.create.reshape import long_averages_to_wide_averages
from datacode.ff.create.minus import construct_minus_variables
from datacode.ff.create.inputs import _standardize_custom_args
from datacode.ff.create.mainport import combine_main_portfolios

def create_ff_factors(df: pd.DataFrame, factor_model: StrOrInt,
                      id_var: str='PERMNO', datevar='Date', byvars: ListOrStr=None,
                      retvar: str='RET', wtvar: str='Market Equity',
                      size_var: str = None, value_var: str = None,
                      profitability_var: str = None, investment_var: str = None,
                      custom_labels: DictofStrsandStrLists = None,
                      custom_groupvar_ngroups_dict: GroupvarNgroupsDict=None,
                      custom_pairings: TwoStrTupleList=None,
                      custom_low_minus_high_dict: StrBoolDict=None) -> pd.DataFrame:

    #### Argument preparation #####

    _validate_model(
        factor_model,
        custom_labels=custom_labels,
        custom_groupvar_ngroups_dict=custom_groupvar_ngroups_dict,
        custom_pairings=custom_pairings,
        custom_low_minus_high_dict=custom_low_minus_high_dict
    )

    custom_labels, custom_groupvar_ngroups_dict, custom_pairings, custom_low_minus_high_dict = _standardize_custom_args(
        custom_labels=custom_labels,
        custom_groupvar_ngroups_dict=custom_groupvar_ngroups_dict,
        custom_pairings=custom_pairings,
        custom_low_minus_high_dict=custom_low_minus_high_dict
    )

    byvars = _to_list_if_str(byvars)

    default_varnames = dict(
        size_var=size_var,
        value_var=value_var,
        profitability_var=profitability_var,
        investment_var=investment_var
    )
    # Convert to portfolio names
    default_portfolio_varnames = {
        key: _other_groupvar_portname(value) if value else value for key, value in default_varnames.items()
    }

    ##### Main logic ##########

    # Assigns portfolios, leaves data in the original shape
    ff_portfolios = create_ff_portfolios(
        df,
        factor_model=factor_model,
        byvars=byvars,
        id_var=id_var,
        date_var=datevar,
        custom_groupvar_ngroups_dict=custom_groupvar_ngroups_dict,
        **default_varnames
    )

    # Replace numbered portfolios with labeled portfolios
    labels = get_and_set_labels(
        ff_portfolios,
        factor_model=factor_model,
        custom_labels=custom_labels,
        **default_portfolio_varnames
    )

    # Get necessary dual sort pairings, create dual sort portfolio variables
    pairings = create_dual_sort_variables_get_pairings(
        ff_portfolios,
        factor_model=factor_model,
        custom_pairings=custom_pairings,
        **default_portfolio_varnames
    )

    if byvars is not None:
        base_vars = byvars + [datevar]
    else:
        base_vars = [datevar]

    # Fama-French portfolio difference procedure. Reduces down to size of time/byvars
    # With 3 factor model, this is the last calculation step, as the SMB factor calculated
    # using value portfolios is the entire SMB factor. Just need to rename after.
    base_df = ff_portfolios.loc[:,base_vars].drop_duplicates()
    for pairing in pairings:
        minus_vars_df = construct_averges_and_minus_variables_for_pairing(
            df=ff_portfolios,
            pairing=pairing,
            labels=labels,
            factor_model=factor_model,
            datevar=datevar,
            byvars=byvars,
            retvar=retvar,
            wtvar=wtvar,
            custom_low_minus_high_dict=custom_low_minus_high_dict,
            **default_portfolio_varnames
        )
        base_df = base_df.merge(minus_vars_df, how='left', on=base_vars)

    # 3 factor model, just rename SMB_HML to SMB
    # For 5 factor model, SMB is calculated separatetely with value ports, profitability
    # ports, and investment ports, then the three are averaged to get the final SMB fator
    combine_main_portfolios(
        df=base_df,
        labels=labels,
        pairings=pairings,
        factor_model=factor_model,
        custom_low_minus_high_dict=custom_low_minus_high_dict,
        **default_portfolio_varnames
    )

    # Add market returns
    mkt_df: pd.DataFrame = market_returns(
        df=ff_portfolios,
        retvar=retvar,
        datevar=datevar,
        wtvar=wtvar,
        byvars=byvars
    )
    base_df = base_df.merge(mkt_df, how='left', on=base_vars)

    return base_df


def construct_averges_and_minus_variables_for_pairing(df: pd.DataFrame, pairing: TwoStrTuple,
                                                      labels: DictofStrsandStrLists, factor_model: StrOrInt,
                                                      datevar='Date', byvars: ListOrStr=None,
                                                      retvar: str='RET', wtvar: str='Market Equity',
                                                      size_var: str = None, value_var: str = None,
                                                      profitability_var: str = None, investment_var: str = None,
                                                      custom_low_minus_high_dict: StrBoolDict=None) -> pd.DataFrame:

    dual_portname = _dual_sort_varname(pairing[0], pairing[1])

    averages_long_shape = portfolio_returns(
        df,
        retvar=retvar,
        datevar=datevar,
        wtvar=wtvar,
        byvars=byvars,
        portvar=dual_portname
    )

    wide = long_averages_to_wide_averages(
        averages_long_shape,
        datevar=datevar,
        retvar=retvar,
        byvars=byvars,
        dual_portvar=dual_portname
    )

    return construct_minus_variables(
        wide,
        labels=labels,
        pairing=pairing,
        factor_model=factor_model,
        byvars=byvars,
        datevar=datevar,
        size_var=size_var,
        value_var=value_var,
        profitability_var=profitability_var,
        investment_var=investment_var,
        custom_low_minus_high_dict=custom_low_minus_high_dict
    )



from datacode.models.dtypes.base import DataType


class BooleanType(DataType):
    names = ('bool', 'bool_')

    def __init__(self, categorical: bool = False, ordered: bool = False):
        super().__init__(
            bool,
            categorical=categorical,
            ordered=ordered
        )

    @classmethod
    def from_str(cls, dtype: str, categorical: bool = False, ordered: bool = False):
        dtype = dtype.lower()
        if dtype not in cls.names:
            raise ValueError(f'Dtype {dtype} does not match valid names for {cls.__name__}: {cls.names}')
        return cls(
            categorical=categorical,
            ordered=ordered
        )



from typing import Sequence, List, Dict, Optional

import pandas as pd
import numpy as np
from IPython.core.display import display, HTML
from semopy import Optimizer, inspect, gather_statistics
import pyexlatex as pl

from datacode.models.variables import Variable
from datacode.sem.eqs import model_eqs
from datacode.sem.to_tex import summary_latex_table
from datacode.sem.constants import STANDARD_SCALE_MESSAGE, ROBUST_SCALE_MESSAGE
from datacode.utils.suppress import no_stdout


def structural_summary_dfs(opt: Optimizer, observed_endog_vars: Sequence[Variable],
                           all_vars: Sequence[Variable]) -> List[pd.DataFrame]:
    summ_dfs = _structural_summary_dfs(opt)
    rename_dict = {var.unique_key: var.name for var in all_vars}
    for i, summ_df in enumerate(summ_dfs):
        summ_df.index.name = observed_endog_vars[i].name
        summ_df.rename(index=rename_dict, inplace=True)

    return summ_dfs


def _structural_summary_dfs(opt: Optimizer) -> List[pd.DataFrame]:
    return _param_summary_dfs(opt, opt.model.vars['ObsEndo'], operator='~')


def latent_summary_dfs(opt: Optimizer, measurement_dict: Dict[Variable, Sequence[Variable]],
                       all_vars: Sequence[Variable]) -> List[pd.DataFrame]:
    summ_dfs = _param_summary_dfs(opt, opt.model.vars['Latents'], operator='=~')
    rename_dict = {var.unique_key: var.name for var in all_vars}
    final_summ_dfs = []
    for summ_df in summ_dfs:
        # Need to add the missing 1 coefficient to the df
        measure_var_key = summ_df.index.name
        # Get the variable which is missing
        missing_var = None
        measure_var = None
        for lhs_var, rhs_vars in measurement_dict.items():
            if lhs_var.unique_key == measure_var_key:
                measure_var = lhs_var
                missing_var = rhs_vars[0]
        if missing_var is None or measure_var is None:
            raise ValueError(f'could not find {measure_var_key} in measurement dict {measurement_dict}')
        missing_row = pd.DataFrame(
            pd.Series([1] + [np.nan] * (len(summ_df.columns) - 1), name=missing_var.name, index=summ_df.columns)
        ).T
        final_summ_df = pd.concat([missing_row, summ_df])
        final_summ_df.index.name = measure_var.name
        final_summ_df.rename(index=rename_dict, inplace=True)
        final_summ_dfs.append(final_summ_df)
    return final_summ_dfs


def _param_summary_dfs(opt: Optimizer, var_keys: Sequence[str], operator: str = '~',
                       include_same_var: bool = False) -> List[pd.DataFrame]:
    with no_stdout():
        results = inspect(opt)
    summ_dfs = []
    for var_key in var_keys:
        selected_results_mask = (results['lval'] == var_key) & (results['op'] == operator)
        if not include_same_var:
            selected_results_mask = selected_results_mask & (results['rval'] != var_key)
        selected_results = results.loc[selected_results_mask]
        summary = selected_results.drop(['lval', 'op'], axis=1).rename(columns={'Value': 'Coef.'}).set_index('rval')
        summary.index.name = var_key
        summ_dfs.append(summary)
    return summ_dfs


def correlation_dfs(opt: Optimizer, all_vars: Sequence[Variable]) -> List[pd.DataFrame]:
    summ_dfs = _correlation_dfs(opt)
    rename_dict = {var.unique_key: var.name for var in all_vars}
    for i, summ_df in enumerate(summ_dfs):
        summ_df.index.name = rename_dict[summ_df.index.name]
        summ_df.rename(index=rename_dict, inplace=True)

    return summ_dfs


def _correlation_dfs(opt: Optimizer) -> List[pd.DataFrame]:
    return _param_summary_dfs(opt, opt.model.vars['All'], operator='~~', include_same_var=True)


def get_fit_statistics(opt: Optimizer) -> pd.DataFrame:
    stats = gather_statistics(opt)
    stat_data = {
        'Model $\chi^2$': stats.chi2[0],
        'Model $p$-value ($\chi^2$)': stats.chi2[1],
        'Baseline $\chi^2$': stats.chi2_baseline,
        'Degrees of Freedom': stats.dof,
        'Goodness of Fit Index (GFI)': stats.gfi,
        'Adjusted-GFI (AGFI)': stats.agfi,
        'Tucker-Lewis Index (TLI)': stats.nfi,
        'Comparative Fit Index (CFI)': stats.cfi,
        'Root Mean Square Error of Approximation (RMSEA)': stats.rmsea
    }
    df = pd.DataFrame(pd.Series(stat_data).apply(lambda x: f'{x:.3f}'), columns=['Statistics'])
    df.loc['Degrees of Freedom'] = stats.dof  # go back to integer
    return df


class SEMSummary:

    def __init__(self, opt: Optimizer, observed_endog_vars: Sequence[Variable],
                 structural_dict: Dict[Variable, Sequence[Variable]],
                 measurement_dict: Dict[Variable, Sequence[Variable]],
                 var_corr_groups: Sequence[Sequence[Variable]],
                 all_vars: Sequence[Variable], scale: bool,
                 robust_scale: bool):
        self.opt = opt
        self.observed_endog_vars = observed_endog_vars
        self.structural_dict = structural_dict
        self.measurement_dict = measurement_dict
        self.var_corr_groups = var_corr_groups
        self.all_vars = all_vars
        self.stats = gather_statistics(opt)
        self.scale = scale
        self.robust_scale = robust_scale

    @property
    def structural(self) -> List[pd.DataFrame]:
        return structural_summary_dfs(self.opt, self.observed_endog_vars, self.all_vars)

    @property
    def latent(self) -> List[pd.DataFrame]:
        return latent_summary_dfs(self.opt, self.measurement_dict, self.all_vars)

    @property
    def correlations(self) -> List[pd.DataFrame]:
        return correlation_dfs(self.opt, self.all_vars)

    @property
    def fit(self) -> pd.DataFrame:
        return get_fit_statistics(self.opt)

    @property
    def eqs(self) -> List[pl.Equation]:
        return model_eqs(self.structural_dict, self.measurement_dict, self.var_corr_groups)

    @property
    def warnings(self) -> List[str]:
        warns = []
        try:
            if not self.opt._sample_cov_pos_df:
                warns.append('Sample covariance matrix is not positive-definite')
        except AttributeError:
            # must be opt created without data-code
            pass

        try:
            if not self.opt._model_cov_pos_df:
                warns.append('Model-implied covariance matrix is not positive-definite')
        except AttributeError:
            # must be opt created without data-code
            pass
        return warns

    @property
    def notes(self) -> List[str]:
        ns = []
        if self.scale:
            ns.append(STANDARD_SCALE_MESSAGE)
        if self.robust_scale:
            ns.append(ROBUST_SCALE_MESSAGE)
        return ns

    def to_latex(self, begin_below_text: Optional[str] = None,
                 end_below_text: Optional[str] = None, replace_below_text: Optional[str] = None,
                 caption: str = 'Structural Equation Model (SEM)') -> pl.Table:
        return summary_latex_table(
            self.fit,
            self.structural,
            self.latent,
            self.scale,
            self.robust_scale,
            equations=self.eqs,
            begin_below_text=begin_below_text,
            end_below_text=end_below_text,
            replace_below_text=replace_below_text,
            caption=caption
        )

    def summary(self):
        if self.warnings:
            _display_header('Warnings')
            _display_unordered_list(self.warnings)
        display(self.fit)
        to_display = {
            'Structural': self.structural,
            'Measurement': self.latent,
            'Correlations': self.correlations,
        }
        for header, df_list in to_display.items():
            _display_header(header)
            for df in df_list:
                display(df)
        if self.notes:
            _display_header('Notes')
            _display_unordered_list(self.notes)

    def _repr_html_(self):
        self.summary()


def _display_header(text: str, level: int = 4):
    display(HTML(f'<h{level}>{text}</h{level}>'))


def _display_unordered_list(items: Sequence[str]):
    if items:
        display(HTML(_ul_html(items)))


def _ul_html(items: Sequence[str]) -> str:
    html_str = '<ul>\n'
    for item in items:
        html_str += f'\t<li>{item}</li>\n'
    html_str += '</ul>'
    return html_str